# encoding=utf-8
from __future__ import unicode_literals


class JobLoggerBase(object):
    """
    Base-class for JobLogger and DummyJobLogger.
    Not part of public API.
    """
    def __init__(self, name, parallel=False, print_to_console=False):
        from .Config import Config
        self.config = Config()
        self._name = name
        self._log_lines = []
        self._error_lines = []
        self._context = []
        self._allow_parallel = parallel
        self._print_to_console = print_to_console or self.config.print_to_console

    @property
    def name(self):
        return self._name

    @property
    def allow_parallel(self):
        return self._allow_parallel

    @property
    def print_to_console(self):
        return self._print_to_console

    @property
    def context(self):
        """Returns the current context-prefix as str"""
        ctx = []
        for c in self._context:
            if not ctx or ctx[-1] != c:
                ctx.append(c)
        ret = ":".join(ctx)
        if ret:
            ret += ": "
        return ret

    def push_context(self, context_name):
        self._context.append("%s" % context_name)

    def pop_context(self):
        self._context = self._context[:-1]



# Generated by Django 2.2 on 2019-04-15 16:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_joblog', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='joblogmodel',
            name='state',
            field=models.CharField(choices=[('running', '▶ running'), ('finished', '✔ finished'), ('error', '❌ error'), ('halted', '❎ halted')], db_index=True, default='running', editable=False, max_length=64, verbose_name='state'),
        ),
    ]



# encoding=utf-8
from __future__ import unicode_literals


from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from django_joblog.models import JobLogModel
from django_joblog import Config, JobLogger


class Command(BaseCommand):
    help = "Set running jobs that seem to be killed to 'vanished' state - for 'ping' mode"

    def add_arguments(self, parser):
        parser.add_argument("-f", "--force", nargs="?", type=bool, const=True, default=False,
                            help="Update the database even though the JOBLOG_CONFIG has 'ping' not enabled")

    def handle(self, *args, **options):
        config = Config()
        if not config.ping:
            if not options.get("force"):
                print("'ping' mode is not enabled in JOBLOG_CONFIG - this may alter the state of running jobs\n"
                      "Use -f/--force to override this warning")
                exit(-1)

        with JobLogger("joblog_cleanup", print_to_console=True):
            JobLogModel.cleanup()



from .impl.Config import Config
from .impl.JobLogger import JobLogger
from .impl.DummyJobLogger import DummyJobLogger
from .impl.JobLoggerContext import JobLoggerContext
from .impl.exceptions import JobIsAlreadyRunningError



import time
from threading import Thread

from django.test import TestCase

from django_joblog.models import JobLogModel, db_alias
from django_joblog import *


class JobLogParallelTestCase(TestCase):

    databases = ("default", "joblog")

    def setUp(self):
        pass

    def test_parallel(self):
        def _task():
            with JobLogger("test-parallel", parallel=True) as log:
                time.sleep(1)
                log.log("Great!")

        threads = []
        for i in range(3):
            threads.append(Thread(target=_task))
            threads[-1].start()

        for t in threads:
            t.join()

        self.assertEqual(3, JobLogModel.objects.using(db_alias()).filter(name="test-parallel").count())

    def test_no_parallel(self):
        def _task():
            with JobLogger("test-no-parallel") as log:
                time.sleep(1)
                log.log("Great!")

        thread = Thread(target=_task)
        thread.start()
        time.sleep(0.1)

        try:
            with self.assertRaises(JobIsAlreadyRunningError):
                with JobLogger("test-no-parallel") as log:
                    log.log("try to run in parallel")

            self.assertEqual(1, JobLogModel.objects.using(db_alias()).filter(name="test-no-parallel").count())

        finally:
            thread.join()



# Generated by Django 2.0.7 on 2018-07-31 16:34

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='JobLogModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, default='', editable=False, max_length=128, verbose_name='name')),
                ('count', models.BigIntegerField(default=0, editable=False, verbose_name='count')),
                ('date_started', models.DateTimeField(db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='started')),
                ('date_ended', models.DateTimeField(blank=True, db_index=True, default=None, editable=False, null=True, verbose_name='ended')),
                ('duration', models.DurationField(blank=True, db_index=True, default=None, null=True, verbose_name='duration')),
                ('state', models.CharField(choices=[('running', '▶ running'), ('finished', '✔ finished'), ('error', '❌ error')], db_index=True, default='running', editable=False, max_length=64, verbose_name='state')),
                ('log_text', models.TextField(blank=True, default=None, editable=False, null=True, verbose_name='log')),
                ('error_text', models.TextField(blank=True, default=None, editable=False, null=True, verbose_name='error log')),
            ],
            options={
                'verbose_name_plural': 'Job logs',
                'verbose_name': 'Job log',
            },
        ),
    ]



# encoding=utf-8
from __future__ import unicode_literals


from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.utils.translation import ugettext_lazy as _

from django_joblog.models import JobLogModel, JobLogStates, db_alias


class Command(BaseCommand):
    help = "Show a single job"

    def add_arguments(self, parser):
        parser.add_argument("pk", type=int,
                            help="Primary key of job to display")

    def handle(self, *args, **options):
        try:
            job = JobLogModel.objects.using(db_alias()).get(pk=options["pk"])

            for n in (
                    "name", "count", "date_started", "date_ended", "duration", "state",
            ):
                print("%12s: %s" % (n, getattr(job, n)))

            if job.log_text:
                print("log:")
                print(job.log_text)
            if job.error_text:
                print("error:")
                print(job.error_text)

        except JobLogModel.DoesNotExist:
            print("Unknown pk '%s'" % options["pk"])



import time
from threading import Thread

from django.test import TestCase
from django.db import transaction

from django_joblog.models import JobLogModel
from django_joblog import *


class JobLogDbUpdatesTestCase(TestCase):

    databases = ("default", "joblog")

    def setUp(self):
        pass

    def test_log_update(self):
        JOB_NAME = "test-log-update"
        MESSAGES = ["one second", "two seconds", "three seconds"]

        def _task():
            with JobLogger(JOB_NAME) as log:
                for msg in MESSAGES:
                    time.sleep(.5)
                    log.log(msg)
                    log.error(msg)

        thread = Thread(target=_task)
        thread.start()
        time.sleep(0.1)

        try:
            self.assertEqual(1, JobLogModel.objects.filter(name=JOB_NAME).count())

            expected_log = []
            for i, msg in enumerate(MESSAGES):
                time.sleep(.5)
                model = JobLogModel.objects.get(name=JOB_NAME)

                expected_log.append(msg)
                self.assertEqual("\n".join(expected_log), model.log_text)
                self.assertEqual("\n".join(expected_log), model.error_text)

        finally:
            thread.join()

    def test_log_update_transaction(self):
        """
        Make sure that a transaction inside a job does not block live-updates of logs
        """
        JOB_NAME = "test-log-update-transaction"
        MESSAGES = ["one second", "two seconds", "three seconds"]

        def _task():
            with JobLogger(JOB_NAME) as log:
                with transaction.atomic():
                    for msg in MESSAGES:
                        time.sleep(.5)
                        log.log(msg)
                        log.error(msg)

        thread = Thread(target=_task)
        thread.start()
        time.sleep(0.1)
        try:
            self.assertEqual(1, JobLogModel.objects.filter(name=JOB_NAME).count())

            expected_log = []
            for i, msg in enumerate(MESSAGES):
                time.sleep(.5)
                model = JobLogModel.objects.get(name=JOB_NAME)

                expected_log.append(msg)
                self.assertEqual("\n".join(expected_log), model.log_text)
                self.assertEqual("\n".join(expected_log), model.error_text)
        finally:
            thread.join()



# encoding=utf-8
from __future__ import unicode_literals

from django.contrib import admin
from django.utils.html import mark_safe, escape
from django.utils.translation import ugettext_lazy as _

from .models import JobLogModel, JobLogStates


class JobLogModelAdmin(admin.ModelAdmin):
    exclude = tuple()
    list_display = (
        "id_decorator", "name", "count",
        "date_started_decorator", "date_ended_decorator", "duration",
        "state_decorator", "log_decorator", "error_log_decorator"
    )
    search_fields = ("name", "log_text", "error_text")
    list_filter = ("name",)

    def id_decorator(self, model):
        return "#%s" % model.id
    id_decorator.short_description = _("ID")

    def date_started_decorator(self, model):
        return model.date_started.strftime("%Y/%m/%d %H:%M:%S")
    date_started_decorator.short_description = _("started")

    def date_ended_decorator(self, model):
        if model.date_ended is None:
            return "-"
        return model.date_ended.strftime("%Y/%m/%d %H:%M:%S")
    date_ended_decorator.short_description = _("ended")

    def state_decorator(self, model):
        state = getattr(JobLogStates, model.state, None)
        if state:
            state = state.value
        else:
            state = model.state
        return mark_safe('<span style="white-space: nowrap">%s</span>' % state)
    state_decorator.short_description = _("state")

    def log_decorator(self, model):
        return _linebreaks(model.log_text, False)
    log_decorator.short_description = _("log")

    def error_log_decorator(self, model):
        return _linebreaks(model.error_text)
    error_log_decorator.short_description = _("error log")


admin.site.register(JobLogModel, JobLogModelAdmin)


def _linebreaks(text, wrap_lines=True):
    if not text:
        return "-"
    lines = []
    for line in text.split("\n"):
        # make html-compatible
        line = escape(line)
        # keep leading whitespace
        for i, c in enumerate(line):
            if not c.isspace() and i:
                line = line[:i].replace(" ", "&nbsp;").replace("\t", "&nbsp;&nbsp;") + line[i:]
                break
        if not wrap_lines:
            line = '<span style="white-space: nowrap">%s</span>' % line
        lines.append(line)
    return mark_safe("<br>".join(lines))



# encoding=utf-8
from __future__ import unicode_literals

import traceback

from django.utils.translation import ugettext_lazy as _

from .JobLoggerBase import JobLoggerBase
from .JobModelAbstraction import JobModelAbstraction


class JobLogger(JobLoggerBase):
    """
    Automatic database logging,
    also catches exceptions and stores the traceback to the database

    with JobLogger("my_task") as job:
        the_task()
        job.log("some_info")
    """
    def __init__(self, name, parallel=False, print_to_console=False):
        """
        Constructs a JobLogger object. Always use with `with` statement.
        :param name: str, The name of the job
        :param parallel: bool, Allow parallel execution of jobs. If False, a JobIsAlreadyRunningError will be
                         thrown, if a job with the same name is actually running
        :param print_to_console: bool, if True, all log, error and exception texts are also printed to the console
        """
        super(JobLogger, self).__init__(name, parallel=parallel, print_to_console=print_to_console)
        self._model = None
        self._thread = None

    def __enter__(self):
        if self._model is None:
            self._model = JobModelAbstraction(self)
            self._model.create_model()
            if self.config.ping:
                from .JobLoggerPingThread import JobLoggerPingThread
                self._thread = JobLoggerPingThread(self)
                self._thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._model is None:
            return True

        try:
            self._model.update_model()
            if exc_tb:
                try:
                    type_name = exc_type.__name__
                except AttributeError:
                    type_name = exc_type
                if exc_val:
                    type_name = "%s - %s" % (type_name, exc_val)
                exc_val = "%s%s\n%s" % (
                    self.context,
                    type_name,
                    "".join(reversed(traceback.format_tb(exc_tb)))
                )
            self._model.finish(exc_val)

            if self._thread is not None:
                self._thread.stop()

            return True

        except BaseException as e:
            self._model.finish("%s: %s" % (e.__class__.__name__, e))

            if self._thread is not None:
                self._thread.stop()

            raise e

    def log(self, line):
        """
        Add a line to the log output
        :param line: anything
        :return: None
        """
        line = self.context + ("%s" % line).strip()
        self._log_lines.append(line)

        if self._model and self.config.live_updates:
            self._model.update_model()

        if self.print_to_console:
            try:
                print("LOG: %s" % line)
            except (UnicodeDecodeError, UnicodeEncodeError):
                pass

    def error(self, line):
        """
        Add a line to the error-output
        :param line: anything
        :return: None
        """
        line = self.context + ("%s" % line).strip()
        self._error_lines.append(line)

        if self._model and self.config.live_updates:
            self._model.update_model()

        if self.print_to_console:
            try:
                print("ERR: %s" % line)
            except (UnicodeDecodeError, UnicodeEncodeError):
                pass




# encoding=utf-8
from __future__ import unicode_literals

from .JobLoggerBase import JobLoggerBase


class DummyJobLogger(JobLoggerBase):
    """
    A class that mimics a JobLogger,
    for functions that require a JobLogger-liker object.

    It provides no __enter__ nor __exit__ functions and can therefore not be
    used with the `with`-statement.

    It simply dumps logging and error output to the console

    with JobLogger("my_task") as log:
        the_task(log)

    def the_task(log=None):
        log = log or DummyJobLogger()
        log.log("some info")
        if 1 != 0:
            log.error("some error")

    """
    def __init__(self, name="", parallel=False):
        super(DummyJobLogger, self).__init__(name, parallel=parallel, print_to_console=True)

    def log(self, line):
        """
        Print text to the console
        :param line: anything
        :return: None
        """
        line = self.context + "%s" % line
        if self.name:
            line = self.name + ": " + line
        try:
            print("LOG: %s" % line)
        except (UnicodeDecodeError, UnicodeEncodeError):
            pass

    def error(self, line):
        """
        Print text to the console
        :param line: anything
        :return: None
        """
        line = self.context + "%s" % line
        if self.name:
            line = self.name + ": " + line
        try:
            print("ERR: %s" % line)
        except (UnicodeDecodeError, UnicodeEncodeError):
            pass


class JobLoggerContext(object):
    """
    Scoped context change for a Logger instance

    with JobLogger("my_task") as log:
        with JobLoggerContext(log, "subtask_1"):
            the_task_1(log)
        with JobLoggerContext(log, "subtask_2"):
            the_task_2(log)
            with JobLoggerContext(log, "subsubtask_2"):
                the_sub_task_2(log)
    """
    def __init__(self, log, name):
        self._log = log
        self._name = name

    def __enter__(self):
        self._log.push_context(self._name)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_tb:
            self._log.pop_context()




# encoding=utf-8
from __future__ import unicode_literals


from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.utils.translation import ugettext_lazy as _

from django_joblog.models import JobLogModel, JobLogStates, db_alias


class Command(BaseCommand):
    help = "List jobs"

    def add_arguments(self, parser):
        parser.add_argument("-s", "--state", nargs="+", type=str, default=[],
                            help="Specify states of jobs to list (running, finished, error, vanished)")

    def handle(self, *args, **options):
        if not options["state"]:
            qset = JobLogModel.objects.using(db_alias()).all()
        else:
            qset = None
            for state in options["state"]:
                state_qset = JobLogModel.objects.using(db_alias()).filter(state=state)
                if qset is None:
                    qset = state_qset
                else:
                    qset |= state_qset

        qset = qset.order_by("-date_started")
        qset = qset[:20]

        format_str = "%4s | %34s | %34s | %30s | %20s"
        print(format_str % ("pk", "date_started", "date_ended", "name", "state"))
        for job in qset:
            print(format_str % (job.pk, job.date_started, job.date_ended, job.name, job.state))



from django.conf import settings


class Config:

    def __init__(self):
        self._config = getattr(settings, "JOBLOG_CONFIG", {})

    @property
    def print_to_console(self):
        return self._config.get("print_to_console", False)

    @property
    def db_alias(self):
        from django.db import DEFAULT_DB_ALIAS
        return self._config.get("db_alias", DEFAULT_DB_ALIAS)

    @property
    def live_updates(self):
        return self._config.get("live_updates", False)

    @property
    def ping(self):
        return self._config.get("ping", False)

    @property
    def ping_interval(self):
        return self._config.get("ping_interval", 10)




# encoding=utf-8
from __future__ import unicode_literals

import threading

from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from .exceptions import JobIsAlreadyRunningError


class JobModelAbstraction(object):
    """
    Helper for JobLogger communication with database.
    Not part of public API.
    """
    def __init__(self, joblog):
        self._p = joblog
        self._job_model = None
        self._model_pk = None
        self._thread = None
        self._update_lock = threading.Lock()

    @property
    def manager(self):
        from django_joblog.models import JobLogModel, db_alias
        return JobLogModel.objects.using(db_alias())

    def is_job_running(self, time_delta=None):
        """
        Return True if a job with the given name is currently running.
        :param time_delta: datetime.timedelta, optional,
                           if not None, return True only if the running job is started within now - time_delta
        :return: bool
        """
        from django_joblog.models import JobLogModel
        return JobLogModel.is_job_running(self._p.name, time_delta=time_delta)

    def create_model(self):
        if not self._p.allow_parallel and self.is_job_running():
            raise JobIsAlreadyRunningError(
                _("The job '%s' is already running and 'parallel' was set to False") % self._p.name
            )

        self._job_model = self._create_model()

    def update_model(self, allow_fail=False):
        from django_joblog.models import db_alias
        with self._update_lock:
            model = self._get_model(allow_fail=allow_fail)
            if not model and allow_fail:
                return

            now = timezone.now()
            duration = now - model.date_started

            log_text = None
            error_text = None

            if self._p._log_lines:
                log_text = "\n".join(self._p._log_lines)

            if self._p._error_lines:
                error_text = "\n".join(self._p._error_lines)

            model.log_text = log_text
            model.error_text = error_text
            model.duration = duration
            model.save(using=db_alias())

    def finish(self, error_text=None):
        if self._model_pk is not None:
            self._finish(error_text)

    def _create_model(self):
        now = timezone.now()
        count = self.manager.filter(name=self._p.name).count() + 1
        if self._p.print_to_console:
            print("\n%s.%s started @ %s" % (self._p.name, count, now))
        model = self.manager.create(name=self._p.name, count=count, date_started=now)
        self._model_pk = model.pk
        return model

    def _get_model(self, allow_fail=False):
        from django_joblog.models import JobLogModel
        if self._model_pk is None:
            return self._create_model()
        else:
            if not allow_fail:
                return self.manager.get(pk=self._model_pk)
            else:
                try:
                    return self.manager.get(pk=self._model_pk)
                except JobLogModel.DoesNotExist:
                    return None

    def _finish(self, exception_or_error=None):
        from django_joblog.models import db_alias, JobLogStates
        
        model = self._get_model()
        
        model.date_ended = timezone.now()
        model.duration = model.date_ended - model.date_started
        model.state = JobLogStates.finished.name
        if exception_or_error is not None:
            if model.error_text:
                model.error_text = "%s\n%s" % (model.error_text, exception_or_error)
            else:
                model.error_text = "%s" % exception_or_error
            model.state = JobLogStates.error.name
        model.save(using=db_alias())

        if self._p.print_to_console:
            if model.log_text or model.error_text:
                print("\n---summary---")
            if model.log_text:
                print("LOG:\n%s" % model.log_text)
            if model.error_text:
                print("ERROR:\n%s" % model.error_text)
            print("%s.%s finished after %s" % (model.name, model.count, model.duration))




# encoding=utf-8
from __future__ import unicode_literals

import time
import threading
import datetime


class JobLoggerPingThread(object):

    """
    Not part of public API
    Helper class to house a thread that regularly calls JobModelAbstraction.update_model()
    """

    def __init__(self, joblog):
        self._p = joblog
        self._thread = None
        self._ping_interval = self._p.config.ping_interval
        self._next_ping_time = None
        self._stop = False

    def start(self):
        if not self._thread:
            self._next_ping_time = datetime.datetime.now() + datetime.timedelta(seconds=self._ping_interval)
            self._thread = threading.Thread(target=self._mainloop)
            self._thread.start()

    def stop(self):
        self._stop = True
        if self._thread:
            self._thread.join()
            self._thread = None

    def _mainloop(self):
        while not self._stop:
            time.sleep(.5)
            if datetime.datetime.now() >= self._next_ping_time:
                self._next_ping_time = datetime.datetime.now() + datetime.timedelta(seconds=self._ping_interval)
                # print("PING")
                if not self._stop:
                    self._p._model.update_model(allow_fail=True)



# encoding=utf-8
from __future__ import unicode_literals

import warnings
import enum

from django.db import models, transaction
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone
from django.conf import settings
from django.db import DEFAULT_DB_ALIAS

from . import Config


class JobLogStates(enum.Enum):
    """
    Enum for all possible states of JobLogModel.state field
    """
    running = _("▶ running")
    finished = _("✔ finished")
    error = _("❌ error")
    vanished = _("⊙ vanished")


def db_alias():
    """
    Returns the db name to use for JobLogModel.
    Return "joblog" if such a database is configured else defaults to django.db.DEFAULT_DB_ALIAS
    :return: str
    """
    alias = Config().db_alias
    if alias not in settings.DATABASES:
        warnings.warn("Configured job-logger db alias '%s' is not in settings.DATABASES" % alias)
        return DEFAULT_DB_ALIAS
    return alias


class JobLogModel(models.Model):
    class Meta:
        verbose_name = _("Job log")
        verbose_name_plural = _("Job logs")

    name = models.CharField(verbose_name=_("name"), max_length=128, editable=False, default="", db_index=True)
    count = models.BigIntegerField(verbose_name=_("count"), editable=False, default=0)
    date_started = models.DateTimeField(verbose_name=_("started"), default=timezone.now, editable=False, db_index=True)
    date_ended = models.DateTimeField(verbose_name=_("ended"), default=None, editable=False, null=True, blank=True, db_index=True)
    duration = models.DurationField(verbose_name=_("duration"), default=None, null=True, blank=True, db_index=True)
    state = models.CharField(verbose_name=_("state"), max_length=64, editable=False, db_index=True,
                             choices=[(e.name, e.value) for e in JobLogStates], default=JobLogStates.running.name)
    log_text = models.TextField(verbose_name=_("log"), default=None, null=True, blank=True, editable=False)
    error_text = models.TextField(verbose_name=_("error log"), default=None, null=True, blank=True, editable=False)

    @classmethod
    def is_job_running(cls, name, time_delta=None, ping=None):
        """
        Return True if a job with the given name is currently running.
        :param name: str
        :param time_delta: datetime.timedelta, optional,
                           if not None, return True only if the running job is started within now - time_delta
        :param ping: bool, overrides the "ping" setting in django.conf.settings.JOBLOG_CONFIG
                When true, jobs will be considered "vanished" when their "duration" field is significantly
                smaller than their true duration
        :return: bool
        """
        if ping is None:
            ping = Config().ping

        if time_delta is None:
            qset = cls.objects.using(db_alias()).filter(
                name=name,
                state=JobLogStates.running.name,
            )
        else:
            date_started = timezone.now() - time_delta
            qset = cls.objects.using(db_alias()).filter(
                name=name,
                state=JobLogStates.running.name,
                date_started__gte=date_started,
            )

        if not qset.exists():
            return False

        if not ping:
            return qset.exists()

        leeway = Config().ping_interval
        now = timezone.now()
        for date_started, duration in qset.values_list("date_started", "duration"):
            true_duration = now - date_started

            if not duration:
                if true_duration.total_seconds() < leeway + 1:
                    # a probably still running job which has not updated it's "duration" field yet
                    return True
                continue

            if true_duration.total_seconds() <= duration.total_seconds() + leeway:
                # a running job who's "duration" has been updated recently
                return True

        return False

    @classmethod
    def cleanup(cls, joblog=None):
        """
        Set all jobs that are running but who's update is older than JOBLOG_CONFIG["ping_interval"] to 'vanished'.
        Careful! JOBLOG_CONFIG["ping"] must be enabled for this to work reliably.
        :param joblog: optional JobLogger instance to log the result
        """
        from django_joblog import DummyJobLogger
        now = timezone.now()
        leeway = Config().ping_interval
        joblog = joblog or DummyJobLogger()

        with transaction.atomic():
            qset = cls.objects.using(db_alias()).filter(state=JobLogStates.running.name)
            num_running = qset.count()
            to_delete = []

            for pk, date_started, duration in qset.values_list("pk", "date_started", "duration"):
                true_duration = now - date_started

                if not duration:
                    if true_duration.total_seconds() < leeway:
                        # a probably still running job which has not updated it's "duration" field yet
                        continue
                else:
                    if true_duration.total_seconds() <= duration.total_seconds() + leeway:
                        # a running job who's "duration" has been updated recently
                        continue

                to_delete.append(pk)

            if not to_delete:
                joblog.log("%s job(s) running, all valid" % num_running)
            else:
                joblog.log("%s job(s) running, %s vanished" % (num_running, len(to_delete)))
                for pk in to_delete:
                    try:
                        job = cls.objects.using(db_alias()).get(pk=pk)
                        job.state = JobLogStates.vanished.name
                        if not job.date_ended:
                            if job.duration:
                                job.date_ended = job.date_started + job.duration
                            else:
                                job.date_ended = now
                        job.save(using=db_alias())
                    except cls.DoesNotExist:
                        pass



# encoding=utf-8
from __future__ import unicode_literals

from .JobLoggerBase import JobLoggerBase


class JobLoggerContext(object):
    """
    Scoped context change for a Logger instance

    with JobLogger("my_task") as log:
        with JobLoggerContext(log, "subtask_1"):
            the_task_1(log)
        with JobLoggerContext(log, "subtask_2"):
            the_task_2(log)
            with JobLoggerContext(log, "subsubtask_2"):
                the_sub_task_2(log)
    """
    def __init__(self, log, name):
        assert isinstance(log, JobLoggerBase)
        self._log = log
        self._name = name

    def __enter__(self):
        self._log.push_context(self._name)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_tb:
            self._log.pop_context()




from setuptools import setup


long_description = "%s\n\n## Changelog\n%s" % (
    open("./README.md").read(),
    open("./HISTORY.md").read(),
)


setup(
    name='django-joblog',
    version='0.2.1',
    description='A generic django-utility that helps to log stuff to the database.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/defgsus/django-joblog',
    author='Stefan Berke',
    author_email='s.berke@netzkolchose.de',
    license='MIT',
    packages=[
        'django_joblog',
        'django_joblog.migrations',
        'django_joblog.management',
        'django_joblog.management.commands',
        'django_joblog.tests',
        'django_joblog.impl',
    ],
    zip_safe=False,
    install_requires=['django>=1.10.0'],
    python_requires='>=2.6, <4',
    keywords="django database logging",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import time, datetime
from threading import Thread

from django.test import TestCase
from django.utils import timezone

from django_joblog.models import JobLogModel, db_alias
from django_joblog import *


def manager():
    return JobLogModel.objects.using(db_alias())


class JobLogBasicTestCase(TestCase):

    databases = ("default", "joblog")

    def setUp(self):
        pass

    def test_db_creation(self):
        with JobLogger("test-db-creation"):
            pass

        qset = manager().filter(name="test-db-creation")
        self.assertEqual(1, qset.count())
        self.assertEqual(1, qset[0].count)
        self.assertEqual("test-db-creation", qset[0].name)

    def test_count(self):
        with JobLogger("test-count"):
            pass
        with JobLogger("test-count"):
            pass

        qset = manager().filter(name="test-count").order_by("-date_started")
        self.assertEqual(2, qset.count())
        self.assertEqual(2, qset[0].count)

    def test_time(self):
        now = timezone.now()
        with JobLogger("test-time"):
            time.sleep(1)

        model = manager().get(name="test-time")

        self.assertLessEqual((model.date_started - now).total_seconds(), 0.01)
        self.assertGreaterEqual(model.duration.total_seconds(), 1.0)
        self.assertLessEqual(model.duration.total_seconds(), 1.6)

    def test_log(self):
        with JobLogger("test-log") as log:
            log.log("Hello!")
            log.log("World!")
            log.error("Hey!")
            log.error("Jude!")

        model = manager().get(name="test-log")

        self.assertEqual("Hello!\nWorld!", model.log_text)
        self.assertEqual("Hey!\nJude!", model.error_text)

    def test_context(self):
        with JobLogger("test-context") as log:
            log.log("outside")
            log.error("eoutside")
            with JobLoggerContext(log, "context1"):
                log.log("inside")
                log.error("einside")
                with JobLoggerContext(log, "context2"):
                    log.log("nested")
                    log.error("enested")
            log.log("outside")
            log.error("eoutside")

        model = manager().get(name="test-context")

        self.assertEqual("outside\ncontext1: inside\ncontext1:context2: nested\noutside", model.log_text)
        self.assertEqual("eoutside\ncontext1: einside\ncontext1:context2: enested\neoutside", model.error_text)

    def test_exception(self):
        with JobLogger("test-exception"):
            raise ValueError("This was bad")

        model = manager().get(name="test-exception")

        self.assertEqual("ValueError - This was bad", model.error_text.split("\n")[0])

    def test_context_exception(self):
        with JobLogger("test-context-exception") as log:
            with JobLoggerContext(log, "context1"):
                with JobLoggerContext(log, "context2"):
                    raise ValueError("This was bad")

        model = manager().get(name="test-context-exception")

        self.assertEqual("context1:context2: ValueError - This was bad", model.error_text.split("\n")[0])

    def test_is_running(self):
        def _task():
            with JobLogger("test-is-running"):
                time.sleep(5)

        thread = Thread(target=_task)
        thread.start()
        time.sleep(.1)

        self.assertEqual(False, JobLogModel.is_job_running("test-is-not-running"))
        self.assertEqual(True, JobLogModel.is_job_running("test-is-running"))
        self.assertEqual(True, JobLogModel.is_job_running("test-is-running", timezone.timedelta(seconds=5)))
        self.assertEqual(True, JobLogModel.is_job_running("test-is-running", datetime.timedelta(seconds=5)))
        time.sleep(1)
        self.assertEqual(False, JobLogModel.is_job_running("test-is-running", timezone.timedelta(seconds=1)))
        self.assertEqual(True, JobLogModel.is_job_running("test-is-running", timezone.timedelta(seconds=3)))

        thread.join()

        self.assertEqual(False, JobLogModel.is_job_running("test-is-running"))
